// Scintilla source code edit control
// Editor.cxx - main code for the edit control
// Copyright 1998-1999 by Neil Hodgson <neilh@hare.net.au>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "Platform.h"

#include "Scintilla.h"
#include "CellBuffer.h"
#include "KeyMap.h"
#include "Indicator.h"
#include "LineMarker.h"
#include "Style.h"
#include "ViewStyle.h"
#include "Document.h"
#include "Editor.h"

Caret::Caret() :
active(true), on(true), period(500) {}

Timer::Timer() :
ticking(false), ticksToWait(0), tickerID(0) {}

Editor::Editor() {
	ctrlID = 0;

	stylesValid = false;

	hideSelection = false;
	inOverstrike = false;
	tabWidth = doc.tabInChars * 8;

	bufferedDraw = true;

	lastClickTime = 0;
	ptMouseLast.x = 0;
	ptMouseLast.y = 0;
	firstExpose = true;
	inDragDrop = false;
	dropWentOutside = false;
	posDrag = -1;
	posDrop = -1;
	selectionType = selChar;

	lastXChosen = 0;
	lineAnchor = 0;
	originalAnchorPos = 0;

	dragChars = 0;
	lenDrag = 0;

	ucWheelScrollLines = 0;
	cWheelDelta = 0;  //wheel delta from roll

	xOffset = 0;
	currentPos = 0;
	anchor = 0;

	topLine = 0;

	doc.AddWatcher(this, 0);
}

Editor::~Editor() {
	DropGraphics();

	delete []dragChars;
	dragChars = 0;
	lenDrag = 0;
}

void Editor::Finalise() {}


void Editor::DropGraphics() {
	pixmapLine.Release();
	pixmapSelMargin.Release();
	pixmapSelPattern.Release();
}

void Editor::InvalidateStyleData() {
	stylesValid = false;
	palette.Release();
	DropGraphics();
}

void Editor::InvalidateStyleRedraw() {
	InvalidateStyleData();
	Redraw();
}

void Editor::RefreshColourPalette(Palette &pal, bool want) {
	vs.RefreshColourPalette(pal, want);
}

void Editor::RefreshStyleData() {
	if (!stylesValid) {
		stylesValid = true;
		vs.Refresh();
		RefreshColourPalette(palette, true);
		palette.Allocate(wMain);
		RefreshColourPalette(palette, false);
		tabWidth = vs.spaceWidth * doc.tabInChars;
		SetScrollBars();
	}
}

PRectangle Editor::GetClientRectangle() {
	return wDraw.GetClientPosition();
}

PRectangle Editor::GetTextRect() {
	PRectangle rc = GetClientRectangle();
	rc.left += vs.fixedColumnWidth;
	return rc;
}

int Editor::LinesOnScreen() {
	PRectangle rcClient = GetClientRectangle();
	int htClient = rcClient.bottom - rcClient.top;
	htClient -= Platform::HeightScrollBar();
	//Platform::DebugPrintf("lines on screen = %d\n", htClient / lineHeight + 1);
	return htClient / vs.lineHeight;
}

int Editor::LinesToScroll() {
	int retVal = LinesOnScreen() - 1;
	if (retVal < 1)
		return 1;
	else
		return retVal;
}

int Editor::MaxScrollPos() {
	//Platform::DebugPrintf("Lines %d screen = %d maxScroll = %d\n",
	//LinesTotal(), LinesOnScreen(), LinesTotal() - LinesOnScreen() + 1);
	//int retVal = LinesTotal() - LinesOnScreen() + 1;
	int retVal = doc.LinesTotal() - LinesOnScreen();
	if (retVal < 0)
		return 0;
	else
		return retVal;
}

Point Editor::LocationFromPosition(unsigned int pos) {
	RefreshStyleData();
	Point pt;
	int line = doc.LineFromPosition(pos);
	//Platform::DebugPrintf("line=%d\n", line);
	int xpos = vs.fixedColumnWidth;
	Surface surface;
	surface.Init();
	pt.y = (line - topLine) * vs.lineHeight; 	// + half a lineheight?
	unsigned int posLineStart = doc.LineStart(line);
	unsigned int posLineEnd = doc.LineStart(line + 1);
	unsigned int selStart = SelectionStart();
	unsigned int selEnd = SelectionEnd();
	int retPos = posLineStart;
	int sty = 0;
	char segment[maxLineLength];
	segment[0] = '\0';
	int segPos = 0;
	char chPrev = '\0';
	bool complete = false;
	//Platform::DebugPrintf("LocationFromPosition A %d %d\n", pos, xpos);
	for (unsigned int i = posLineStart;
	        i <= posLineEnd && !complete && (posLineEnd > posLineStart);
	        i++) {
		char ch = doc.CharAt(i);
		int colour = doc.StyleAt(i);
		unsigned int width = 0;
		if (colour != sty || (ch == '\t') || (chPrev == '\t') || (i == selStart) || (i == selEnd) || (i == posLineEnd)) {
			if (segment[0] == '\t') {
				width = (((xpos + 2 - vs.fixedColumnWidth) / tabWidth) + 1)
				        * tabWidth + vs.fixedColumnWidth - xpos;
				if (i >= pos) {
					xpos += width;
					complete = true;
				}
			} else {
				int styleMain = sty & 31;
				if (i >= pos) {
					int charsToMeasure = segPos - (i - pos);
					if (charsToMeasure > 0) {
						width = surface.WidthText(vs.styles[styleMain].font, segment, charsToMeasure);
						xpos += width;
					}
					complete = true;
				} else {
					width = surface.WidthText(vs.styles[styleMain].font, segment, segPos);
				}
			}
			if (!complete) {
				retPos += segPos;
				segPos = 0;
				segment[segPos] = '\0';
				xpos += width;
				sty = colour;
			}
		}
		if (ch != '\r' && ch != '\n' && (static_cast<unsigned int>(segPos + 1) < sizeof(segment)))
			segment[segPos++] = ch;
		chPrev = ch;
	}
	pt.x = xpos - xOffset;
	return pt;
}

int Editor::LineFromLocation(Point pt) {
	return pt.y / vs.lineHeight + topLine;
}

int Editor::PositionFromLocation(Point pt) {
	pt.x += xOffset;
	int line = pt.y / vs.lineHeight + topLine;
	if (pt.y < 0)	// Division rounds towards 0
		line = (pt.y - (vs.lineHeight - 1)) / vs.lineHeight + topLine;
	if (line < 0)
		return 0;
	if (line >= doc.LinesTotal())
		return doc.Length();
	//Platform::DebugPrintf("Position of (%d,%d) line = %d top=%d\n", pt.x, pt.y, line, topLine);
	int xpos = vs.fixedColumnWidth;
	Surface surface;
	surface.Init();
	unsigned int posLineStart = doc.LineStart(line);
	unsigned int posLineEnd = doc.LineStart(line + 1);
	unsigned int selStart = SelectionStart();
	unsigned int selEnd = SelectionEnd();
	unsigned int retPos = posLineStart;
	int sty = 0;
	char segment[maxLineLength];
	segment[0] = '\0';
	unsigned int segPos = 0;
	char chPrev = '\0';
	bool complete = false;
	for (unsigned int i = posLineStart;
	        i <= posLineEnd && xpos < pt.x && !complete && (posLineEnd > posLineStart);
	        i++) {
		char ch = doc.CharAt(i);
		int colour = doc.StyleAt(i);
		int width = 0;
		if (colour != sty || (ch == '\t') || (chPrev == '\t') || (i == selStart) ||
		        (i == selEnd) || (i == posLineEnd)) {
			if (segment[0] == '\t') {
				width = (((xpos + 2 - vs.fixedColumnWidth) / tabWidth) + 1)
				        * tabWidth + vs.fixedColumnWidth - xpos;
				if ((xpos + width) >= pt.x) {
					complete = true;
					if ((xpos + width / 2) <= pt.x)	// If in first half of char
						retPos = retPos + 1;
				}
			} else {
				int styleMain = sty & 31;
				width = surface.WidthText(vs.styles[styleMain].font, segment, segPos);
				if ((xpos + width) >= pt.x) {
					// For non -tabs must measure positions
					int prevWidth = 0;
					for (unsigned int j = 1; j <= segPos && !complete; j++) {
						width = surface.WidthText(vs.styles[styleMain].font, segment, j);
						if ((xpos + width) >= pt.x) {
							complete = true;
							retPos += j - 1;
							if ((xpos + prevWidth + (width - prevWidth) / 2) < pt.x)
								retPos = retPos + 1;
						}
						prevWidth = width;
					}
				}
			}
			if (!complete) {
				retPos += segPos;
				segPos = 0;
				segment[segPos] = '\0';
				xpos += width;
				sty = colour;
			}
		}
		if (ch != '\r' && ch != '\n' && ((segPos + 1) < sizeof(segment)))
			segment[segPos++] = ch;
		chPrev = ch;
	}

	if (retPos > posLineEnd)
		retPos = posLineEnd;
	if (retPos < posLineStart)
		retPos = posLineStart;
	return retPos;
}

void Editor::RedrawRect(PRectangle rc) {
	//Platform::DebugPrintf("Redraw %d %d - %d %d\n", rc->left, rc->top, rc->right, rc->bottom);
	wDraw.InvalidateRectangle(rc);
}

void Editor::Redraw() {
	//Platform::DebugPrintf("Redraw all\n");
	wDraw.InvalidateAll();
}

void Editor::RedrawSelMargin() {
	if (vs.fixedColumnWidth > 0) {
		PRectangle rcSelMargin = GetClientRectangle();
		rcSelMargin.right = vs.fixedColumnWidth;
		wDraw.InvalidateRectangle(rcSelMargin);
	} else {
		Redraw();
	}
}

void Editor::InvalidateRange(int start, int end) {
	int minPos = start;
	if (minPos > end)
		minPos = end;
	int maxPos = start;
	if (maxPos < end)
		maxPos = end;
	int minLine = doc.LineFromPosition(minPos);
	int maxLine = doc.LineFromPosition(maxPos);
	PRectangle rcClient = GetTextRect();
	PRectangle rcRedraw;
	rcRedraw.left = vs.fixedColumnWidth;
	rcRedraw.top = (minLine - topLine) * vs.lineHeight;
	rcRedraw.right = rcClient.right;
	rcRedraw.bottom = (maxLine - topLine + 1) * vs.lineHeight;
	// Ensure PRectangle is within 16 bit space
	rcRedraw.top = Platform::Clamp(rcRedraw.top, -32000, 32000);
	rcRedraw.bottom = Platform::Clamp(rcRedraw.bottom, -32000, 32000);

	RedrawRect(rcRedraw);
}

int Editor::CurrentPosition() {
	return currentPos;
}

int Editor::SelectionStart() {
	return Platform::Minimum(currentPos, anchor);
}

int Editor::SelectionEnd() {
	return Platform::Maximum(currentPos, anchor);
}

void Editor::SetSelection(int currentPos_, int anchor_) {
	currentPos_ = doc.ClampPositionIntoDocument(currentPos_);
	anchor_ = doc.ClampPositionIntoDocument(anchor_);
	if ((currentPos != currentPos_) || (anchor != anchor_)) {
		int firstAffected = anchor;
		if (firstAffected > currentPos)
			firstAffected = currentPos;
		if (firstAffected > anchor_)
			firstAffected = anchor_;
		if (firstAffected > currentPos_)
			firstAffected = currentPos_;
		int lastAffected = anchor;
		if (lastAffected < currentPos)
			lastAffected = currentPos;
		if (lastAffected < anchor_)
			lastAffected = anchor_;
		if (lastAffected < (currentPos_ + 1))	// +1 ensures caret repainted
			lastAffected = (currentPos_ + 1);
		currentPos = currentPos_;
		anchor = anchor_;
		InvalidateRange(firstAffected, lastAffected);
	}
	ClaimSelection();
}

void Editor::SetSelection(int currentPos_) {
	currentPos_ = doc.ClampPositionIntoDocument(currentPos_);
	if (currentPos != currentPos_) {
		int firstAffected = anchor;
		if (firstAffected > currentPos)
			firstAffected = currentPos;
		if (firstAffected > currentPos_)
			firstAffected = currentPos_;
		int lastAffected = anchor;
		if (lastAffected < currentPos)
			lastAffected = currentPos;
		if (lastAffected < (currentPos_ + 1))	// +1 ensures caret repainted
			lastAffected = (currentPos_ + 1);
		currentPos = currentPos_;
		InvalidateRange(firstAffected, lastAffected);
	}
	ClaimSelection();
}

void Editor::SetEmptySelection(int currentPos_) {
	SetSelection(currentPos_, currentPos_);
}

void Editor::SetPosition(int pos, bool shift) {
	int oldPos = currentPos;
	currentPos = doc.ClampPositionIntoDocument(pos);
	currentPos = doc.MovePositionOutsideChar(currentPos, oldPos - currentPos);
	if (!shift)
		anchor = currentPos;
	EnsureCaretVisible();
	ClaimSelection();
}

int Editor::MovePositionTo(int newPos, bool extend) {
	int delta = newPos - currentPos;
	newPos = doc.ClampPositionIntoDocument(newPos);
	newPos = doc.MovePositionOutsideChar(newPos, delta);
	if (extend) {
		SetSelection(newPos);
	} else {
		SetEmptySelection(newPos);
	}
	EnsureCaretVisible();
	ShowCaretAtCurrentPosition();
	return 0;
}

// Choose the x position that the caret will try to stick to as it is moves up and down
void Editor::SetLastXChosen() {
	Point pt = LocationFromPosition(currentPos);
	lastXChosen = pt.x;
}

void Editor::ScrollTo(int line) {
	topLine = Platform::Clamp(line, 0, MaxScrollPos());
	SetVerticalScrollPos();
	Redraw();
}

void Editor::HorizontalScrollTo(int xPos) {
	//Platform::DebugPrintf("HorizontalScroll %d\n", xPos);
	xOffset = xPos;
	if (xOffset < 0)
		xOffset = 0;
	SetHorizontalScrollPos();
	Redraw();
}

void Editor::EnsureCaretVisible() {
	//Platform::DebugPrintf("EnsureCaretVisible %d\n", xOffset);
	PRectangle rcClient = GetTextRect();
	int posCaret = currentPos;
	if (posDrag >= 0)
		posCaret = posDrag;
	Point pt = LocationFromPosition(posCaret);
	Point ptBottomCaret = pt;
	int lineCaret = doc.LineFromPosition(posCaret);
	ptBottomCaret.y += vs.lineHeight - 1;
	if (!rcClient.Contains(pt) || !rcClient.Contains(ptBottomCaret)) {
		//Platform::DebugPrintf("EnsureCaretVisible move, (%d,%d) (%d,%d)\n", pt.x, pt.y, rcClient.left, rcClient.right);
		if (topLine > lineCaret) {
			ScrollTo(lineCaret);
		} else if (topLine < (lineCaret - LinesOnScreen())) {
			ScrollTo(lineCaret - LinesOnScreen());
		}
		// The 2s here are to ensure the caret is really visible
		if (pt.x < rcClient.left) {
			xOffset = xOffset - (rcClient.left - pt.x) - 2;
		} else if (pt.x > rcClient.right) {
			xOffset = xOffset + (pt.x - rcClient.right) + 2;
		}
		if (xOffset < 0)
			xOffset = 0;
		SetHorizontalScrollPos();
		Redraw();
	}
}

void Editor::ShowCaretAtCurrentPosition() {
	if (!wMain.HasFocus()) {
		caret.active = false;
		caret.on = false;
		return;
	}
	caret.active = true;
	caret.on = true;
	SetTicking(true);
}

void Editor::DropCaret() {
	caret.active = false;
	InvalidateCaret();
}

void Editor::InvalidateCaret() {
	if (posDrag >= 0)
		InvalidateRange(posDrag, posDrag + 1);
	else
		InvalidateRange(currentPos, currentPos + 1);
}

void Editor::PaintSelMargin(Surface *surfWindow, PRectangle &rc) {
	if (vs.fixedColumnWidth == 0)
		return;

	PRectangle rcMargin = GetClientRectangle();
	rcMargin.right = vs.fixedColumnWidth;

	PRectangle rcSelMargin = rcMargin;
	rcSelMargin.left = vs.lineNumberWidth;

	if (!rc.Intersects(rcMargin))
		return;

	Surface *surface;
	if (bufferedDraw) {
		surface = &pixmapSelMargin;
	} else {
		surface = surfWindow;
	}

	// Required because of special way brush is created for selection margin
	surface->FillRectangle(rcSelMargin, pixmapSelPattern); {	// Scope the line and yposScreen variables
		int line = topLine;
		int yposScreen = 0;

		while (line < doc.LinesTotal() && yposScreen < rcMargin.bottom) {
			int marks = doc.GetMark(line);
			if (marks) {
				PRectangle rcMarker;
				rcMarker.left = 1 + vs.lineNumberWidth;
				rcMarker.top = yposScreen + 1;
				rcMarker.right = vs.lineNumberWidth + vs.selMarginWidth - 1;
				rcMarker.bottom = yposScreen + vs.lineHeight - 1;
				for (int markBit = 0; (markBit < 32) && marks; markBit++) {
					if (marks & 1) {
						vs.markers[markBit].Draw(surface, rcMarker);
					}
					marks >>= 1;
				}
			}
			line++;
			yposScreen += vs.lineHeight;
		}
	}

	if (vs.lineNumberWidth > 0) {
		int line = topLine;
		int ypos = 0;

		while (ypos < rcMargin.bottom) {
			char number[100];
			number[0] = '\0';
			if (line < doc.LinesTotal())
				sprintf(number, "%d", line + 1);
			int xpos = 0;
			PRectangle rcNumber;
			rcNumber.left = xpos;
			rcNumber.right = xpos + vs.lineNumberWidth;
			rcNumber.top = ypos;
			rcNumber.bottom = ypos + vs.lineHeight;
			surface->FillRectangle(rcNumber, vs.selbar.allocated);
			// Right justify
			int width = surface->WidthText(vs.styles[0].font, number, strlen(number));
			xpos += vs.lineNumberWidth - width - 3;
			rcNumber.left = xpos;
			surface->DrawText(rcNumber, vs.styles[0].font,
			                  ypos + vs.maxAscent, number, strlen(number),
			                  vs.linenumfore.allocated, vs.selbar.allocated);
			line++;
			ypos += vs.lineHeight;
		}
	}

	if (bufferedDraw) {
		surfWindow->Copy(rcMargin, Point(), pixmapSelMargin);
	}
}

void Editor::Paint(Surface *surfaceWindow, PRectangle rcPaint) {
	//Platform::DebugPrintf("Paint %d %d - %d %d\n", rcPaint.left, rcPaint.top, rcPaint.right, rcPaint.bottom);
	RefreshStyleData();

	PRectangle rcClient = GetClientRectangle();
	//Platform::DebugPrintf("Client: (%3d,%3d) ... (%3d,%3d)   %d\n",
	//	rcClient.left, rcClient.top, rcClient.right, rcClient.bottom);

	if (!pixmapSelPattern.Initialised()) {
		pixmapSelPattern.InitPixMap(8, 8, surfaceWindow);
		// This complex procedure is to reproduce the checker board dithered pattern used by windows
		// for scroll bars and Visual Studio for its selection margin. The colour of this pattern is half
		// way between the chrome colour and the chrome highlight colour making a nice transition
		// between the window chrome and the content area. And it works in low colour depths.
		PRectangle rcPattern(0, 0, 8, 8);
		if (vs.selbarlight.desired == Colour(0xff, 0xff, 0xff)) {
			pixmapSelPattern.FillRectangle(rcPattern, vs.selbar.allocated);
			pixmapSelPattern.PenColour(vs.selbarlight.allocated);
			for (int stripe = 0; stripe < 8; stripe++) {
				pixmapSelPattern.MoveTo(0, stripe * 2);
				pixmapSelPattern.LineTo(8, stripe * 2 - 8);
			}
		} else {
			// User has choen an unusual chrome colour scheme so just use the highlight edge colour.
			pixmapSelPattern.FillRectangle(rcPattern, vs.selbarlight.allocated);
		}
	}

	if (bufferedDraw) {
		if (!pixmapLine.Initialised()) {
			pixmapLine.InitPixMap(rcClient.Width(), rcClient.Height(),
			                      surfaceWindow);
			pixmapSelMargin.InitPixMap(vs.fixedColumnWidth,
			                           rcClient.Height(), surfaceWindow);
		}
	}

	surfaceWindow->SetPalette(&palette, true);
	pixmapLine.SetPalette(&palette, !wMain.HasFocus());

	//Platform::DebugPrintf("Paint: (%3d,%3d) ... (%3d,%3d)   %d\n",
	//	rcPaint.left, rcPaint.top, rcPaint.right, rcPaint.bottom);

	int screenLinePaintFirst = rcPaint.top / vs.lineHeight;
	int linePaintLast = topLine + rcPaint.bottom / vs.lineHeight + 1;
	int endPosPaint = doc.Length();
	if (linePaintLast < doc.LinesTotal())
		endPosPaint = doc.LineStart(linePaintLast + 1)-1;

	if (endPosPaint > doc.GetEndStyled()) {
		// Notify container to do some more styling
		NotifyStyleNeeded(endPosPaint);
	}

	int sty = 0;
	char segment[maxLineLength];
	int xpos = vs.fixedColumnWidth;
	int ypos = 0;
	if (!bufferedDraw)
		ypos += screenLinePaintFirst * vs.lineHeight;
	int yposScreen = screenLinePaintFirst * vs.lineHeight;

	PaintSelMargin(surfaceWindow, rcPaint);

	//Platform::DebugPrintf("start display %d tabWidth = %d, margin = %d offset = %d\n", doc.Length(), tabWidth, selMarginWidth, xOffset);
	int selStart = SelectionStart();
	int selEnd = SelectionEnd();

	Surface *surface = 0;
	if (rcPaint.right > vs.selMarginWidth) {

		if (bufferedDraw) {
			surface = &pixmapLine;
		} else {
			surface = surfaceWindow;
		}

		int line = topLine + screenLinePaintFirst;

		int posCaret = currentPos;
		if (posDrag >= 0)
			posCaret = posDrag;

		// Remove selection margin from drawing area so text will not be drawn
		// on it in unbuffered mode.
		PRectangle rcTextArea = rcClient;
		rcTextArea.left = vs.fixedColumnWidth;
		surfaceWindow->SetClip(rcTextArea);
		while (line < doc.LinesTotal() && yposScreen < rcPaint.bottom) {

			int marks = 0;
			Colour markBack = Colour(0, 0, 0);
			if (vs.selMarginWidth == 0) {
				marks = doc.GetMark(line);
				if (marks) {
					for (int markBit = 0; (markBit < 32) && marks; markBit++) {
						if (marks & 1) {
							markBack = vs.markers[markBit].back.allocated;
						}
						marks >>= 1;
					}
				}
				marks = doc.GetMark(line);
			}

			segment[0] = '\0';
			unsigned int segPos = 0;

			int posLineStart = doc.LineStart(line);
			int posLineEnd = doc.LineStart(line + 1);
			//Platform::DebugPrintf("line %d %d - %d\n", line, posLineStart, posLineEnd);

			PRectangle rcBlank;
			rcBlank.top = ypos;
			rcBlank.bottom = ypos + vs.lineHeight;

			bool inSelection = posLineStart > selStart && posLineStart < selEnd;

			int indicatorsSet = 0;
			if (posLineStart < posLineEnd)
				indicatorsSet = doc.StyleAt(posLineStart) & INDICS_MASK;

			int prevIndic = doc.StyleAt(posLineStart) & INDICS_MASK;

			int indStart[INDIC_MAX + 1] = {0};
			for (int indica = 0; indica <= INDIC_MAX; indica++)
				indStart[indica] = vs.fixedColumnWidth;

			char chPrev = '\0';
			bool visibleInSelection = false;
			int ch = ' ';
			int colour = -1;

			int xposCaret = -1;
			int widthOverstrikeCaret = 1;

			for (int i = posLineStart; i <= posLineEnd; i++) {
				//Platform::DebugPrintf("pos %d\n", line, posLineStart, posLineEnd);
				if (i < posLineEnd) {	// Do not index onto next line
					ch = doc.CharAt(i);
					colour = doc.StyleAt(i);
				}

				// If there is the end of a style run for any reason
				if (colour != sty || (ch == '\t') || (chPrev == '\t') ||
				        (i == selStart) || (i == selEnd) || (i == posCaret) || (i == posLineEnd)) {
					int styleMain = sty & 31;
					// text appears not to have a background color, so draw backing rect first
					Colour textBack = vs.styles[styleMain].back.allocated;
					Colour textFore = vs.styles[styleMain].fore.allocated;
					if (inSelection && !hideSelection) {
						if (vs.selbackset)
							textBack = vs.selbackground.allocated;
						if (vs.selforeset)
							textFore = vs.selforeground.allocated;
						visibleInSelection = true;
					} else {
						if (marks)
							textBack = markBack;
					}
					unsigned int width = 0;
					if (segment[0] == '\t') {
						int nextTab = (((xpos + 2 - vs.fixedColumnWidth) /
						                tabWidth) + 1) * tabWidth +
						              vs.fixedColumnWidth;
						rcBlank.left = xpos - xOffset;
						rcBlank.right = nextTab - xOffset;
						surface->FillRectangle(rcBlank, textBack);
						if (vs.viewWhitespace) {
							PRectangle rcTab;
							rcTab.left = rcBlank.left + 1;
							rcTab.top = rcBlank.top + 4;
							rcTab.right = rcBlank.right - 1;
							rcTab.bottom = rcBlank.bottom - vs.maxDescent;
							int ymid = ypos + vs.lineHeight / 2;
							int ydiff = (rcTab.bottom - rcTab.top) / 2;
							int xhead = rcTab.right - 1 - ydiff;
							surface->PenColour(textFore);
							if ((rcTab.left + 2) < (rcTab.right - 1))
								surface->MoveTo(rcTab.left + 2, ymid);
							else
								surface->MoveTo(rcTab.right - 1, ymid);
							surface->LineTo(rcTab.right - 1, ymid);
							surface->LineTo(xhead, ymid - ydiff);
							surface->MoveTo(rcTab.right - 1, ymid);
							surface->LineTo(xhead, ymid + ydiff);
						}
						width = nextTab - xpos;
					} else {
						//Platform::DebugPrintf("pos %d %d %s\n", line, i, segment);
						width = surface->WidthText(vs.styles[styleMain].font, segment, segPos);
						rcBlank.left = xpos - xOffset;
						rcBlank.right = xpos + width - xOffset;
						rcBlank.top = ypos;
						rcBlank.bottom = ypos + vs.lineHeight;
						surface->DrawText(rcBlank, vs.styles[styleMain].font,
						                  ypos + vs.maxAscent, segment, segPos, textFore, textBack);
						if (vs.viewWhitespace) {
							surface->PenColour(textFore);
							int xx = xpos;
							for (unsigned int cpos = 0; cpos < segPos; cpos++) {
								int szc = surface->WidthChar(vs.styles[styleMain].font, segment[cpos]);
								if (segment[cpos] == ' ') {
									PRectangle rcDot(xx + szc / 2 - xOffset, ypos + vs.lineHeight / 2, 0, 0);
									rcDot.right = rcDot.left + 1;
									rcDot.bottom = rcDot.top + 1;
									surface->FillRectangle(rcDot, textFore);
								}
								xx += szc;
							}
						}
					}

					xpos += width;
					segPos = 0;
					segment[segPos] = '\0';
					sty = colour;

					if (i == selStart)
						inSelection = true;

					if (i == selEnd && (i < posLineEnd))
						inSelection = false;

					if (i == posCaret && (i < posLineEnd || (line == doc.LinesTotal() - 1))) {
						xposCaret = xpos;
						widthOverstrikeCaret =
						    surface->WidthChar(vs.styles[colour & 31].font, ch);
						if (widthOverstrikeCaret < 3)
							widthOverstrikeCaret = 3;
					}

					if (i < posLineEnd) {	// Do not index onto next line
						indicatorsSet = sty & INDICS_MASK;
					}
					if (indicatorsSet != prevIndic) {
						int mask = INDIC0_MASK;
						int indicnum = 0;
						for (indicnum = 0; indicnum <= INDIC_MAX; indicnum++) {
							if ((indicatorsSet & mask) && !(prevIndic & mask)) {
								indStart[indicnum] = xpos;
							}
							mask = mask << 1;
						}

						mask = INDIC0_MASK;
						for (indicnum = 0; indicnum <= INDIC_MAX; indicnum++) {
							if (!(indicatorsSet & mask) && (prevIndic & mask)) {
								PRectangle rcIndic(
								    indStart[indicnum] - xOffset,
								    ypos + vs.maxAscent,
								    xpos - xOffset,
								    ypos + vs.maxAscent + 3);
								vs.indicators[indicnum].Draw(surface, rcIndic);
							}
							mask = mask << 1;
						}
						prevIndic = indicatorsSet;
					}
				}

				if (ch != '\r' && ch != '\n' && ((segPos + 1) < sizeof(segment)))
					segment[segPos++] = ch;
				chPrev = ch;
			}

			// Fill in a PRectangle representing the end of line characters
			rcBlank.left = xpos - xOffset;
			rcBlank.right = xpos + vs.aveCharWidth - xOffset;
			if (inSelection && !hideSelection && visibleInSelection && vs.selbackset &&
			        ((ch == '\r') || (ch == '\n'))) {
				surface->FillRectangle(rcBlank, vs.selbackground.allocated);
			} else if (marks) {
				surface->FillRectangle(rcBlank, markBack);
			} else {
				surface->FillRectangle(rcBlank, vs.styles[sty & 31].back.allocated);
			}

			rcBlank.left = xpos + vs.aveCharWidth - xOffset;
			rcBlank.right = rcClient.right;
			if (marks) {
				surface->FillRectangle(rcBlank, markBack);
			} else if (vs.styles[sty & 31].eolFilled) {
				surface->FillRectangle(rcBlank, vs.styles[sty & 31].back.allocated);
			} else {
				surface->FillRectangle(rcBlank, vs.background.allocated);
			}

			if (((caret.active && caret.on) || (posDrag >= 0)) && xposCaret >= 0) {
				PRectangle rcCaret = rcBlank;
				if (posDrag >= 0) {
					rcCaret.left = xposCaret - xOffset;
					rcCaret.right = xposCaret - xOffset + 1;
				} else {
					if (inOverstrike) {
						rcCaret.top = rcCaret.bottom - 2;
						rcCaret.left = xposCaret - xOffset + 1;
						rcCaret.right = rcCaret.left + widthOverstrikeCaret - 1;
					} else {
						rcCaret.left = xposCaret - xOffset;
						rcCaret.right = xposCaret - xOffset + 1;
					}
				}
				surface->FillRectangle(rcCaret, vs.caretcolour.allocated);
			}

			if (bufferedDraw) {
				Point from(vs.fixedColumnWidth, 0);
				PRectangle rcLine(vs.fixedColumnWidth, yposScreen,
				                  rcClient.right, yposScreen + vs.lineHeight);
				surfaceWindow->Copy(rcLine, from, pixmapLine);
			}

			if (!bufferedDraw) {
				ypos += vs.lineHeight;
			}

			yposScreen += vs.lineHeight;
			xpos = vs.fixedColumnWidth;
			line++;
		}
		PRectangle rcBeyondEOF = rcClient;
		rcBeyondEOF.left = vs.fixedColumnWidth;
		rcBeyondEOF.right = rcBeyondEOF.right;
		rcBeyondEOF.top = (doc.LinesTotal() - topLine) * vs.lineHeight;
		if (rcBeyondEOF.top < rcBeyondEOF.bottom) {
			surfaceWindow->FillRectangle(rcBeyondEOF, vs.background.allocated);
		}
	}
}

// This is mostly copied from the Paint method but with some things omitted
// such as the margin indicators, line numbers, selection and caret
// Should be merged back into a combined Draw method.
long Editor::FormatRange(bool draw, FORMATRANGE *pfr) {
	int printMaxAscent = 1;
	int printMaxDescent = 1;
	int printAveCharWidth = 3;
	Font fonts[STYLE_MAX + 1];
	Surface *surface = new Surface();
	surface->Init(pfr->hdc);
	Surface *surfaceMeasure = new Surface();
	surfaceMeasure->Init(pfr->hdcTarget);
	unsigned int j;
	for (j = 0; j < (sizeof(fonts) / sizeof(fonts[0])); j++) {
		int deviceHeight = (vs.styles[j].size * surfaceMeasure->LogPixelsY()) / 72;
		fonts[j].Create(vs.fontName, deviceHeight,
		                vs.styles[j].bold, vs.styles[j].italic);
		if (j == 0)
			printAveCharWidth = surfaceMeasure->AverageCharWidth(fonts[j]);
		int ascent = surfaceMeasure->Ascent(fonts[j]);
		int descent = surfaceMeasure->Descent(fonts[j]);
		if (printMaxAscent < ascent)
			printMaxAscent = ascent;
		if (printMaxDescent < descent)
			printMaxDescent = descent;
		//Platform::DebugPrintf("Height for font %d = %d\n", i, surface->Height(fonts[i]));
	}



	int printLineHeight = printMaxAscent + printMaxDescent;

	int printSpaceWidth = surfaceMeasure->WidthChar(fonts[0], ' ');
	int printTabWidth = printSpaceWidth * doc.tabInChars;

	int linePrintStart = doc.LineFromPosition(pfr->chrg.cpMin);
	int linePrintLast = linePrintStart + (pfr->rc.bottom - pfr->rc.top) / printLineHeight - 1;
	if (linePrintLast < linePrintStart)
		linePrintLast = linePrintStart;
	int linePrintMax = doc.LineFromPosition(pfr->chrg.cpMax - 1);
	if (linePrintLast > linePrintMax)
		linePrintLast = linePrintMax;
	//Platform::DebugPrintf("Formatting lines=[%0d,%0d,%0d] top=%0d bottom=%0d line=%0d %0d\n",
	//	linePrintStart, linePrintLast, linePrintMax, pfr->rc.top, pfr->rc.bottom, printLineHeight,
	//	surfaceMeasure->Height(fonts[0]));
	int endPosPrint = doc.Length();
	if (linePrintLast < doc.LinesTotal())
		endPosPrint = doc.LineStart(linePrintLast + 1);

	if (endPosPrint > doc.GetEndStyled()) {
		// Notify container to do some more styling
		NotifyStyleNeeded(endPosPrint);
	}
	int sty = 0;
	char segment[maxLineLength];
	int xpos = pfr->rc.left;
	int ypos = pfr->rc.top;
	int line = linePrintStart;

	if (draw) {	// Otherwise just measuring

		while (line <= linePrintLast && ypos < pfr->rc.bottom) {

			int marks = 0;
			Colour markBack = Colour(0, 0, 0);
			if (vs.selMarginWidth == 0) {
				marks = doc.GetMark(line);
				if (marks) {
					for (int markBit = 0; (markBit < 32) && marks; markBit++) {
						if (marks & 1) {
							markBack = vs.markers[markBit].back.allocated;
						}
						marks >>= 1;
					}
				}
				marks = doc.GetMark(line);
			}

			segment[0] = '\0';
			unsigned int segPos = 0;

			int posLineStart = doc.LineStart(line);
			int posLineEnd = doc.LineStart(line + 1);
			//Platform::DebugPrintf("line %d %d - %d\n", line, posLineStart, posLineEnd);

			PRectangle rcBlank;
			rcBlank.top = ypos;
			rcBlank.bottom = ypos + printLineHeight;

			int indicatorsSet = 0;
			if (posLineStart < posLineEnd)
				indicatorsSet = doc.StyleAt(posLineStart) & INDICS_MASK;

			int prevIndic = doc.StyleAt(posLineStart) & INDICS_MASK;

			int indStart[INDIC_MAX + 1] = {0};
			for (int indica = 0; indica <= INDIC_MAX; indica++)
				indStart[indica] = vs.fixedColumnWidth;

			char chPrev = '\0';
			int ch = ' ';
			int colour = -1;

			for (int i = posLineStart; i <= posLineEnd; i++) {
				//Platform::DebugPrintf("pos %d\n", line, posLineStart, posLineEnd);
				if (i < posLineEnd) {	// Do not index onto next line
					ch = doc.CharAt(i);
					colour = doc.StyleAt(i);
				}

				// If there is the end of a style run for any reason
				if (colour != sty || (ch == '\t') || (chPrev == '\t') ||
				        (i == posLineEnd)) {
					int styleMain = sty & 31;
					// text appears not to have a background color, so draw backing rect first
					Colour textBack = vs.styles[styleMain].back.desired;
					Colour textFore = vs.styles[styleMain].fore.desired;
					if (marks)
						textBack = markBack;
					unsigned int width = 0;
					if (segment[0] == '\t') {
						int nextTab = (((xpos + 2) / printTabWidth) + 1) * printTabWidth;
						rcBlank.left = xpos;
						rcBlank.right = nextTab;
						surface->FillRectangle(rcBlank, textBack);
						if (vs.viewWhitespace) {
							PRectangle rcTab;
							rcTab.left = rcBlank.left + 1;
							rcTab.top = rcBlank.top + 4;
							rcTab.right = rcBlank.right - 1;
							rcTab.bottom = rcBlank.bottom - vs.maxDescent;
							int ymid = ypos + printLineHeight / 2;
							int ydiff = (rcTab.bottom - rcTab.top) / 2;
							int xhead = rcTab.right - 1 - ydiff;
							surface->PenColour(textFore);
							if ((rcTab.left + 2) < (rcTab.right - 1))
								surface->MoveTo(rcTab.left + 2, ymid);
							else
								surface->MoveTo(rcTab.right - 1, ymid);
							surface->LineTo(rcTab.right - 1, ymid);
							surface->LineTo(xhead, ymid - ydiff);
							surface->MoveTo(rcTab.right - 1, ymid);
							surface->LineTo(xhead, ymid + ydiff);
						}
						width = nextTab - xpos;
					} else {
						//Platform::DebugPrintf("pos %d %d %s\n", line, i, segment);
						width = surface->WidthText(fonts[styleMain], segment, segPos);
						rcBlank.left = xpos;
						rcBlank.right = xpos + width;
						rcBlank.top = ypos;
						rcBlank.bottom = ypos + printLineHeight;
						surface->DrawText(rcBlank, fonts[styleMain],
						                  ypos + vs.maxAscent, segment, segPos,
						                  textFore, textBack);
						if (vs.viewWhitespace) {
							surface->PenColour(textFore);
							int xx = xpos;
							for (unsigned int cpos = 0; cpos < segPos; cpos++) {
								int szc = surface->WidthChar(fonts[styleMain], segment[cpos]);
								if (segment[cpos] == ' ') {
									PRectangle rcDot(xx + szc / 2, ypos + printLineHeight / 2, 0, 0);
									rcDot.right = rcDot.left + 1;
									rcDot.bottom = rcDot.top + 1;
									surface->FillRectangle(rcDot, textFore);
								}
								xx += szc;
							}
						}
					}

					xpos += width;
					segPos = 0;
					segment[segPos] = '\0';
					sty = colour;

					if (i < posLineEnd) {	// Do not index onto next line
						indicatorsSet = sty & INDICS_MASK;
					}
					if (indicatorsSet != prevIndic) {
						int mask = INDIC0_MASK;
						int indicnum = 0;
						for (indicnum = 0; indicnum <= INDIC_MAX; indicnum++) {
							if ((indicatorsSet & mask) && !(prevIndic & mask)) {
								indStart[indicnum] = xpos;
							}
							mask = mask << 1;
						}

						mask = INDIC0_MASK;
						for (indicnum = 0; indicnum <= INDIC_MAX; indicnum++) {
							if (!(indicatorsSet & mask) && (prevIndic & mask)) {
								PRectangle rcIndic(
								    indStart[indicnum] - xOffset,
								    ypos + vs.maxAscent,
								    xpos - xOffset,
								    ypos + vs.maxAscent + 3);
								vs.indicators[indicnum].Draw(surface, rcIndic);
							}
							mask = mask << 1;
						}
						prevIndic = indicatorsSet;
					}
				}

				if (ch != '\r' && ch != '\n' && ((segPos + 1) < sizeof(segment)))
					segment[segPos++] = ch;
				chPrev = ch;
			}

			// Fill in a PRectangle representing the end of line characters
			rcBlank.left = xpos;
			rcBlank.right = xpos + vs.aveCharWidth;
			if (marks) {
				surface->FillRectangle(rcBlank, markBack);
			} else {
				surface->FillRectangle(rcBlank, vs.styles[sty & 31].back.allocated);
			}

			rcBlank.left = xpos + vs.aveCharWidth - xOffset;
			rcBlank.right = pfr->rc.right;
			if (marks) {
				surface->FillRectangle(rcBlank, markBack);
			} else if (vs.styles[sty & 31].eolFilled) {
				surface->FillRectangle(rcBlank, vs.styles[sty & 31].back.allocated);
			} else {
				surface->FillRectangle(rcBlank, vs.background.allocated);
			}

			ypos += printLineHeight;
			xpos = 0;
			line++;
		}
	}

	delete surface;
	delete surfaceMeasure;

	for (j = 0; j < (sizeof(fonts) / sizeof(fonts[0])); j++) {
		fonts[j].Release();
	}
	return endPosPrint;
}

void Editor::SetScrollBarsTo(PRectangle) {
	RefreshStyleData();

	int nMax = doc.LinesTotal();
	int nPage = doc.LinesTotal() - MaxScrollPos() + 1;
	bool modified = ModifyScrollBars(nMax, nPage);

	// TODO: ensure always showing as many lines as possible
	// May not be, if, for example, window made larger
	if (topLine > MaxScrollPos()) {
		topLine = Platform::Clamp(topLine, 0, MaxScrollPos());
		SetVerticalScrollPos();
		Redraw();
	}
	if (modified)
		Redraw();
	//Platform::DebugPrintf("end max = %d page = %d\n", nMax, nPage);
}




void Editor::SetScrollBars() {
	PRectangle rsClient = GetClientRectangle();
	SetScrollBarsTo(rsClient);
}

void Editor::AddChar(char ch) {
	bool wasSelection = currentPos != anchor;
	ClearSelection();
	if (inOverstrike && !wasSelection) {
		if (currentPos < (doc.Length() - 1)) {
			if ((doc.CharAt(currentPos) != '\r') && (doc.CharAt(currentPos) != '\n')) {
				doc.DelChar(currentPos);
			}
		}
	}
	doc.InsertChar(currentPos, ch);
	SetEmptySelection(currentPos + 1);
	EnsureCaretVisible();
	if (!wasSelection && isprint(ch)) {
		// Try to optimise redraw to just the current line
		// Colourise line and see if line end style changes
		int line = doc.LineFromPosition(currentPos);
		int lineEnd = doc.Length();
		if (line < doc.LinesTotal())
			lineEnd = doc.LineStart(line + 1) - 1;
		int colour = doc.StyleAt(lineEnd);
		// Try to style to end of line including line end characters
		if (lineEnd < doc.Length() - 1)
			NotifyStyleNeeded(lineEnd + 1);
		else
			NotifyStyleNeeded(lineEnd);
		//Platform::DebugPrintf("Addchar trying line = %d lineEnd = %d colour = %d new colour = %d\n",
		//		line, lineEnd, colour,
		//		doc.StyleAt(lineEnd));
		if (colour != doc.StyleAt(lineEnd)) {
			//Platform::DebugPrintf("Fallback addchar %x\n", ch);
			Redraw();
		} else {
			InvalidateRange(doc.LineStart(line), doc.LineEndPosition(currentPos));
			//Platform::DebugPrintf("Optimal %d %d-%d\n", line, LineStart(line), LineEndPosition(currentPos));
		}



	} else {
		//Platform::DebugPrintf("Standard addchar %x\n", ch);
		Redraw();
	}
	NotifyChar(ch);
}

void Editor::ClearSelection() {
	int startPos = SelectionStart();
	unsigned int chars = SelectionEnd() - startPos;
	SetEmptySelection(startPos);
	if (0 != chars) {
		doc.DeleteChars(startPos, chars);
	}
}

void Editor::ClearAll() {
	if (0 != doc.Length()) {
		doc.DeleteChars(0, doc.Length());
	}
	anchor = 0;
	currentPos = 0;
	topLine = 0;
	SetVerticalScrollPos();
	Redraw();
}

void Editor::Cut() {
	Copy();
	ClearSelection();
	Redraw();
}

void Editor::Clear() {
	if (currentPos == anchor) {
		DelChar();
	} else {
		ClearSelection();
	}
	SetEmptySelection(currentPos);
	Redraw();
}

void Editor::SelectAll() {
	SetSelection(0, doc.Length());
	Redraw();
}

void Editor::Undo() {
	if (doc.CanUndo()) {
		int newPos = doc.Undo();
		SetEmptySelection(newPos);
		EnsureCaretVisible();
		Redraw();
		SetScrollBars();
	}
}

void Editor::Redo() {
	if (doc.CanRedo()) {
		int newPos = doc.Redo();
		SetEmptySelection(newPos);
		EnsureCaretVisible();
		Redraw();
		SetScrollBars();
	}
}

void Editor::DelChar() {
	doc.DelChar(currentPos);
	Redraw();
}

void Editor::DelCharBack() {
	if (currentPos == anchor) {
		int newPos = doc.DelCharBack(currentPos);
		SetEmptySelection(newPos);
	} else {
		ClearSelection();
		SetEmptySelection(currentPos);
	}
	Redraw();
}

void Editor::NotifyStyleNeeded(int endStyleNeeded) {
	SCNotification scn;
	scn.nmhdr.code = SCN_STYLENEEDED;
	scn.position = endStyleNeeded;
	NotifyParent(scn);
}

void Editor::NotifyChar(char ch) {
	SCNotification scn;
	scn.nmhdr.code = SCN_CHARADDED;
	scn.ch = ch;
	NotifyParent(scn);
}

void Editor::NotifySavePoint(bool isSavePoint) {
	SCNotification scn;
	if (isSavePoint) {
		scn.nmhdr.code = SCN_SAVEPOINTREACHED;
	} else {
		scn.nmhdr.code = SCN_SAVEPOINTLEFT;
	}
	NotifyParent(scn);
}

void Editor::NotifyModifyAttempt() {
	SCNotification scn;
	scn.nmhdr.code = SCN_MODIFYATTEMPTRO;
	NotifyParent(scn);
}

void Editor::NotifyDoubleClick(Point, bool) {
	SCNotification scn;
	scn.nmhdr.code = SCN_DOUBLECLICK;
	NotifyParent(scn);
}

// Notifications from document
void Editor::NotifyModifyAttempt(Document*, void *) {
	//Platform::DebugPrintf("** Modify Attempt\n");
	NotifyModifyAttempt();
}

void Editor::NotifySavePoint(Document*, void *, bool atSavePoint) {
	//Platform::DebugPrintf("** Save Point %s\n", atSavePoint ? "On" : "Off");
	NotifySavePoint(atSavePoint);
}

void Editor::NotifyModified(Document*, void *) {
	//Platform::DebugPrintf("** Changed\n");
	NotifyChange();
	SetScrollBars();
}

int Editor::KeyCommand(WORD iMessage) {
	Point pt = LocationFromPosition(currentPos);

	switch (iMessage) {
	case SCI_LINEDOWN:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y + vs.lineHeight)));
	case SCI_LINEDOWNEXTEND:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y + vs.lineHeight)), true);
	case SCI_LINEUP:
		MovePositionTo(PositionFromLocation(
		                   Point(lastXChosen, pt.y - vs.lineHeight)));
		return 0;
	case SCI_LINEUPEXTEND:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y - vs.lineHeight)), true);
	case SCI_CHARLEFT:
		MovePositionTo(currentPos - 1);
		SetLastXChosen();
		return 0;
	case SCI_CHARLEFTEXTEND:
		MovePositionTo(currentPos - 1, true);
		SetLastXChosen();
		return 0;
	case SCI_CHARRIGHT:
		MovePositionTo(currentPos + 1);
		SetLastXChosen();
		return 0;
	case SCI_CHARRIGHTEXTEND:
		MovePositionTo(currentPos + 1, true);
		SetLastXChosen();
		return 0;
	case SCI_WORDLEFT:
		return MovePositionTo(doc.NextWordStart(currentPos, -1));
	case SCI_WORDLEFTEXTEND:
		return MovePositionTo(doc.NextWordStart(currentPos, -1), true);
	case SCI_WORDRIGHT:
		return MovePositionTo(doc.NextWordStart(currentPos, 1));
	case SCI_WORDRIGHTEXTEND:
		return MovePositionTo(doc.NextWordStart(currentPos, 1), true);
	case SCI_HOME:
		MovePositionTo(doc.LineStart(doc.LineFromPosition(currentPos)));
		SetLastXChosen();
		return 0;
	case SCI_HOMEEXTEND:
		MovePositionTo(doc.LineStart(doc.LineFromPosition(currentPos)), true);
		SetLastXChosen();
		return 0;
	case SCI_LINEEND:
		MovePositionTo(doc.LineEndPosition(currentPos));
		SetLastXChosen();
		return 0;
	case SCI_LINEENDEXTEND:
		MovePositionTo(doc.LineEndPosition(currentPos), true);
		SetLastXChosen();
		return 0;
	case SCI_DOCUMENTSTART:
		MovePositionTo(0);
		SetLastXChosen();
		return 0;
	case SCI_DOCUMENTSTARTEXTEND:
		MovePositionTo(0, true);
		SetLastXChosen();
		return 0;
	case SCI_DOCUMENTEND:
		MovePositionTo(doc.Length());
		SetLastXChosen();
		return 0;
	case SCI_DOCUMENTENDEXTEND:
		MovePositionTo(doc.Length(), true);
		SetLastXChosen();
		return 0;
	case SCI_PAGEUP:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y - vs.lineHeight * LinesToScroll())));
	case SCI_PAGEUPEXTEND:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y - vs.lineHeight * LinesToScroll())), true);
	case SCI_PAGEDOWN:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y + vs.lineHeight * LinesToScroll())));
	case SCI_PAGEDOWNEXTEND:
		return MovePositionTo(PositionFromLocation(
		                          Point(lastXChosen, pt.y + vs.lineHeight * LinesToScroll())), true);
	case SCI_EDITTOGGLEOVERTYPE:
		inOverstrike = !inOverstrike;
		DropCaret();
		ShowCaretAtCurrentPosition();
		return 0;
	case SCI_CANCEL: 	// Cancel any modes - handled in subclass


		return 0;
	case SCI_DELETEBACK:
		DelCharBack();
		EnsureCaretVisible();
		break;
	case SCI_TAB:
		Indent(true);
		break;
	case SCI_BACKTAB:
		Indent(false);
		break;
	case SCI_NEWLINE:
		ClearSelection();
		if (doc.eolMode == SC_EOL_CRLF) {
			doc.InsertString(currentPos, "\r\n");
			SetEmptySelection(currentPos + 2);
			NotifyChar('\r');
			NotifyChar('\n');
		} else if (doc.eolMode == SC_EOL_CR) {
			doc.InsertChar(currentPos, '\r');
			SetEmptySelection(currentPos + 1);
			NotifyChar('\r');
		} else if (doc.eolMode == SC_EOL_LF) {
			doc.InsertChar(currentPos, '\n');
			SetEmptySelection(currentPos + 1);
			NotifyChar('\n');
		}
		EnsureCaretVisible();
		break;
	case SCI_FORMFEED:
		AddChar('\f');
		break;
	case SCI_VCHOME:
		MovePositionTo(doc.VCHomePosition(currentPos));
		SetLastXChosen();
		return 0;
	case SCI_VCHOMEEXTEND:
		MovePositionTo(doc.VCHomePosition(currentPos), true);
		SetLastXChosen();
		break;
	case SCI_ZOOMIN:
		if (vs.zoomLevel < 20)
			vs.zoomLevel++;
		InvalidateStyleRedraw();
		break;
	case SCI_ZOOMOUT:
		if (vs.zoomLevel > -10)
			vs.zoomLevel--;
		InvalidateStyleRedraw();
		break;
	case SCI_DELWORDLEFT: {
			int startWord = doc.NextWordStart(currentPos, -1);
			doc.DeleteChars(startWord, currentPos - startWord);
			MovePositionTo(startWord);
			Redraw();
		}
		break;
	case SCI_DELWORDRIGHT: {
			int endWord = doc.NextWordStart(currentPos, 1);
			doc.DeleteChars(currentPos, endWord - currentPos);
			Redraw();
		}
		break;
	}
	return 0;
}

int Editor::KeyDown(int key, bool shift, bool ctrl, bool alt) {
	int modifiers = (shift ? SCI_SHIFT : 0) | (ctrl ? SCI_CTRL : 0) |
	                (alt ? SCI_ALT : 0);
	int msg = kmap.Find(key, modifiers);
	if (msg)
		return WndProc(msg, 0, 0);
	else
		return KeyDefault(key, modifiers);
	return TRUE;
}

void Editor::SetWhitespaceVisible(bool view) {
	vs.viewWhitespace = view;
}

bool Editor::GetWhitespaceVisible() {
	return vs.viewWhitespace;
}

void Editor::Indent(bool forwards) {
	//Platform::DebugPrintf("INdent %d\n", forwards);
	int lineOfAnchor = doc.LineFromPosition(anchor);
	int lineCurrentPos = doc.LineFromPosition(currentPos);
	if (lineOfAnchor == lineCurrentPos) {
		ClearSelection();
		doc.InsertChar(currentPos++, '\t');
		SetEmptySelection(currentPos);
	} else {
		int anchorPosOnLine = anchor - doc.LineStart(lineOfAnchor);
		int currentPosPosOnLine = currentPos - doc.LineStart(lineCurrentPos);
		// Multiple lines selected so indent / dedent
		int lineTopSel = Platform::Minimum(lineOfAnchor, lineCurrentPos);
		int lineBottomSel = Platform::Maximum(lineOfAnchor, lineCurrentPos);
		if (doc.LineStart(lineBottomSel) == anchor || doc.LineStart(lineBottomSel) == currentPos)
			lineBottomSel--; 	// If not selecting any characters on a line, do not indent
		doc.BeginUndoAction();
		doc.Indent(forwards, lineBottomSel, lineTopSel);
		doc.EndUndoAction();
		if (lineOfAnchor < lineCurrentPos) {
			if (currentPosPosOnLine == 0)
				SetSelection(doc.LineStart(lineCurrentPos), doc.LineStart(lineOfAnchor));
			else
				SetSelection(doc.LineStart(lineCurrentPos + 1), doc.LineStart(lineOfAnchor));
		} else {
			if (anchorPosOnLine == 0)
				SetSelection(doc.LineStart(lineCurrentPos), doc.LineStart(lineOfAnchor));
			else
				SetSelection(doc.LineStart(lineCurrentPos), doc.LineStart(lineOfAnchor + 1));
		}
	}
}

long Editor::FindText(WORD iMessage, WPARAM wParam, LPARAM lParam) {
	FINDTEXTEX *ft = reinterpret_cast<FINDTEXTEX *>(lParam);
	int pos = doc.FindText(ft->chrg.cpMin, ft->chrg.cpMax, ft->lpstrText,
	                       wParam & FR_MATCHCASE, wParam & FR_WHOLEWORD);
	if (pos != -1) {
		if (iMessage == EM_FINDTEXTEX) {
			ft->chrgText.cpMin = pos;
			ft->chrgText.cpMax = pos + strlen(ft->lpstrText);
		}
	}
	return pos;
}

void Editor::GoToLine(int lineNo) {
	if (lineNo > doc.LinesTotal())
		lineNo = doc.LinesTotal();
	if (lineNo < 0)
		lineNo = 0;
	SetEmptySelection(doc.LineStart(lineNo));
	ShowCaretAtCurrentPosition();
	EnsureCaretVisible();
}

static bool Close(Point pt1, Point pt2) {
	if (abs(pt1.x - pt2.x) > 3)
		return false;
	if (abs(pt1.y - pt2.y) > 3)
		return false;
	return true;
}

char *Editor::CopyRange(int start, int end) {
	char *text = 0;
	if (start < end) {
		int len = end - start;
		text = new char[len + 1];
		if (text) {
			for (int i = 0; i < len; i++) {
				text[i] = doc.CharAt(start + i);
			}
			text[len] = '\0';
		}
	}
	return text;
}

char *Editor::CopySelectionRange() {
	return CopyRange(SelectionStart(), SelectionEnd());
}

void Editor::CopySelectionIntoDrag() {
	delete []dragChars;
	dragChars = 0;
	lenDrag = SelectionEnd() - SelectionStart();
	dragChars = CopyRange(SelectionStart(), SelectionEnd());
	if (!dragChars) {
		lenDrag = 0;
	}
}

void Editor::SetDragPosition(int newPos) {
	if (newPos >= 0) {
		newPos = doc.MovePositionOutsideChar(newPos, 1);
		posDrop = newPos;
	}
	if (posDrag != newPos) {
		caret.on = true;
		SetTicking(true);
		InvalidateCaret();
		posDrag = newPos;
		InvalidateCaret();
	}
}

void Editor::StartDrag() {
	// Always handled by subclasses
	//SetMouseCapture(true);
	//wDraw.SetCursor(Window::cursorArrow);
}

void Editor::DropAt(int position, const char *value, bool moving) {
//Platform::DebugPrintf("DropAt %d\n", inDragDrop);
	int selStart = SelectionStart();
	int selEnd = SelectionEnd();

	if (inDragDrop)
		dropWentOutside = false;

	if ((!inDragDrop) || (position <= selStart) || (position >= selEnd)) {

		doc.BeginUndoAction();

		if (inDragDrop && moving) {
			// Remove dragged out text
			ClearSelection();
			if (position > selStart) {
				position -= selEnd - selStart;
			}
		}

		position = doc.MovePositionOutsideChar(position, currentPos - position);
		doc.InsertString(position, value);

		doc.EndUndoAction();

		SetSelection(position + strlen(value), position);

	} else if (inDragDrop) {
		SetSelection(position, position);
	}
}

void Editor::ButtonDown(Point pt, unsigned int curTime, bool shift) {
	//Platform::DebugPrintf("Scintilla:ButtonDown %d %d = %d\n", curTime, lastClickTime, curTime - lastClickTime);
	ptMouseLast = pt;
	int newPos = PositionFromLocation(pt);
	newPos = doc.MovePositionOutsideChar(newPos, currentPos - newPos);
	inDragDrop = false;
	if (shift) {
		SetSelection(newPos);
	}
	if (((curTime - lastClickTime) < Platform::DoubleClickTime()) && Close(pt, lastClick)) {
		//Platform::DebugPrintf("Double click %d %d = %d\n", curTime, lastClickTime, curTime - lastClickTime);
		SetMouseCapture(true);
		SetEmptySelection(newPos);
		bool doubleClick = false;
		// Stop mouse button bounce changing selection type
		if (curTime != lastClickTime) {
			if (selectionType == selChar) {
				selectionType = selWord;
				doubleClick = true;
			} else if (selectionType == selWord) {
				selectionType = selLine;
			} else {
				selectionType = selChar;
				originalAnchorPos = currentPos;
			}
		}

		if (selectionType == selWord) {
			if (currentPos >= originalAnchorPos) {	// Moved forward
				SetSelection(doc.ExtendWordSelect(currentPos, 1),
				             doc.ExtendWordSelect(originalAnchorPos, -1));
			} else {	// Moved backward
				SetSelection(doc.ExtendWordSelect(currentPos, -1),
				             doc.ExtendWordSelect(originalAnchorPos, 1));
			}
		} else if (selectionType == selLine) {
			lineAnchor = LineFromLocation(pt);
			SetSelection(doc.LineStart(lineAnchor + 1), doc.LineStart(lineAnchor));
			//Platform::DebugPrintf("Triple click: %d - %d\n", anchor, currentPos);
		}
		else {
			SetEmptySelection(currentPos);
		}
		//Platform::DebugPrintf("Double click: %d - %d\n", anchor, currentPos);
		if (doubleClick)
			NotifyDoubleClick(pt, shift);
	} else {
		if (!shift) {
			if (currentPos > anchor) {
				inDragDrop = (newPos >= anchor) && (newPos <= currentPos);
			} else if (currentPos < anchor) {
				inDragDrop = (newPos <= anchor) && (newPos >= currentPos);
			}
		}
		if (inDragDrop) {
			SetMouseCapture(false);
			SetDragPosition(newPos);
			CopySelectionIntoDrag();
			StartDrag();
		} else {
			SetDragPosition( -1);
			SetMouseCapture(true);
			if (!shift)
				SetEmptySelection(newPos);
			selectionType = selChar;
			originalAnchorPos = currentPos;
		}
	}
	lastClickTime = curTime;
	Redraw();
	lastXChosen = pt.x;
	ShowCaretAtCurrentPosition();
}

void Editor::ButtonMove(Point pt) {
	//Platform::DebugPrintf("Move %d %d\n", pt.x, pt.y);
	if (HaveMouseCapture()) {
		ptMouseLast = pt;
		int movePos = PositionFromLocation(pt);
		movePos = doc.MovePositionOutsideChar(movePos, currentPos - movePos);
		if (posDrag >= 0) {
			SetDragPosition(movePos);
		} else {
			if (selectionType == selChar) {
				SetSelection(movePos);
				//Platform::DebugPrintf("Move: %d - %d\n", anchor, currentPos);
			} else if (selectionType == selWord) {
				// Continue selecting by word
				if (currentPos > originalAnchorPos) {	// Moved forward
					SetSelection(doc.ExtendWordSelect(movePos, 1),
					             doc.ExtendWordSelect(originalAnchorPos, -1));
				} else {	// Moved backward
					SetSelection(doc.ExtendWordSelect(movePos, -1),
					             doc.ExtendWordSelect(originalAnchorPos, 1));
				}
			} else {
				// Continue selecting by line
				int lineMove = LineFromLocation(pt);
				if (lineAnchor < lineMove) {
					SetSelection(doc.LineStart(lineMove + 1),
					             doc.LineStart(lineAnchor));
				} else {
					SetSelection(doc.LineStart(lineAnchor + 1),
					             doc.LineStart(lineMove));
				}
			}
		}
		EnsureCaretVisible();
	}
}

void Editor::ButtonUp(Point pt, unsigned int curTime, bool ctrl) {
	//Platform::DebugPrintf("ButtonUp %d\n", HaveMouseCapture());
	if (HaveMouseCapture()) {
		wDraw.SetCursor(Window::cursorText);
		ptMouseLast = pt;
		SetMouseCapture(false);
		int newPos = PositionFromLocation(pt);
		newPos = doc.MovePositionOutsideChar(newPos, currentPos - newPos);
		if (inDragDrop) {
			int selStart = SelectionStart();
			int selEnd = SelectionEnd();
			if (selStart < selEnd) {
				if (dragChars && lenDrag) {
					if (ctrl) {
						doc.InsertString(newPos, dragChars, lenDrag);
						SetSelection(newPos, newPos + lenDrag);
					} else if (newPos < selStart) {
						doc.DeleteChars(selStart, lenDrag);
						doc.InsertString(newPos, dragChars, lenDrag);
						SetSelection(newPos, newPos + lenDrag);
					} else if (newPos > selEnd) {
						doc.DeleteChars(selStart, lenDrag);
						newPos -= lenDrag;
						doc.InsertString(newPos, dragChars, lenDrag);
						SetSelection(newPos, newPos + lenDrag);
					} else {
						SetEmptySelection(newPos);
					}
					delete []dragChars;
					dragChars = 0;
					lenDrag = 0;
				}
				selectionType = selChar;
			}
		} else {
			if (selectionType == selChar) {
				SetSelection(newPos);
			}
		}
		lastClickTime = curTime;
		lastClick = pt;
		lastXChosen = pt.x;
		inDragDrop = false;
	}
}

// Called frequently to perform background UI including
// caret blinking and automatic scrolling.
void Editor::Tick() {
	if (HaveMouseCapture()) {
		// Auto scroll
		ButtonMove(ptMouseLast);
	}
	if (caret.period > 0) {
		timer.ticksToWait -= timer.tickSize;
		if (timer.ticksToWait <= 0) {
			caret.on = !caret.on;
			timer.ticksToWait = caret.period;
			InvalidateCaret();
		}
	}
}

long Editor::WndProc(WORD iMessage, WPARAM wParam, LPARAM lParam) {
	//Platform::DebugPrintf("S start wnd proc %d %d %d\n",iMessage, wParam, lParam);
	switch (iMessage) {

	case WM_GETTEXT:
		{
			char *ptr = reinterpret_cast<char *>(lParam);
			for (unsigned int iChar = 0; iChar < wParam; iChar++)
				ptr[iChar] = doc.CharAt(iChar);
		}
		break;

	case WM_SETTEXT:
		{
			doc.DeleteChars(0, doc.Length());
			SetEmptySelection(0);
			doc.InsertString(0, reinterpret_cast<char *>(lParam));
			Redraw();
		}
		break;

	case WM_GETTEXTLENGTH:
		return doc.Length();

	case WM_NOTIFY:
		//Platform::DebugPrintf("S notify %d %d\n", wParam, lParam);


		break;

	case WM_CUT:
		Cut();
		break;

	case WM_COPY:
		Copy();
		break;

	case WM_PASTE:
		Paste();
		SetScrollBars();
		break;

	case WM_CLEAR:
		//Platform::DebugPrintf("S Clear %d %x %x\n",iMessage, wParam, lParam);


		Clear();
		SetScrollBars();
		break;

	case WM_UNDO:
		Undo();
		break;

		// Edit control mesages

		// Not supported (no-ops):
		//		EM_GETWORDBREAKPROC
		//		EM_GETWORDBREAKPROCEX
		//		EM_SETWORDBREAKPROC
		//		EM_SETWORDBREAKPROCEX
		//		EM_GETWORDWRAPMODE
		//		EM_SETWORDWRAPMODE
		//		EM_LIMITTEXT
		//		EM_EXLIMITTEXT
		//		EM_SETRECT
		//		EM_SETRECTNP
		//		EM_FMTLINES
		//		EM_GETHANDLE
		//		EM_SETHANDLE
		//		EM_GETPASSWORDCHAR
		//		EM_SETPASSWORDCHAR
		//		EM_SETTABSTOPS
		//		EM_FINDWORDBREAK
		//		EM_GETCHARFORMAT
		//		EM_SETCHARFORMAT
		//		EM_GETOLEINTERFACE
		//		EM_SETOLEINTERFACE
		//		EM_SETOLECALLBACK
		//		EM_GETPARAFORMAT
		//		EM_SETPARAFORMAT
		//		EM_PASTESPECIAL
		//		EM_REQUESTRESIZE
		//		EM_GETBKGNDCOLOR
		//		EM_SETBKGNDCOLOR
		//		EM_STREAMIN
		//		EM_STREAMOUT
		//		EM_GETIMECOLOR
		//		EM_SETIMECOLOR
		//		EM_GETIMEOPTIONS
		//		EM_SETIMEOPTIONS
		//		EM_GETMARGINS
		//		EM_SETMARGINS
		//		EM_GETOPTIONS
		//		EM_SETOPTIONS
		//		EM_GETPUNCTUATION
		//		EM_SETPUNCTUATION
		//		EM_GETTHUMB

		// Not supported but should be:
		//		EM_GETEVENTMASK
		//		EM_SETEVENTMASK
		//		For printing:
		//			EM_DISPLAYBAND
		//			EM_SETTARGETDEVICE

	case EM_CANUNDO:
		return doc.CanUndo() ? TRUE : FALSE;

	case EM_UNDO:
		Undo();
		SetScrollBars();
		break;

	case EM_EMPTYUNDOBUFFER:
		doc.DeleteUndoHistory();
		return 0;

	case EM_GETFIRSTVISIBLELINE:
		return topLine;

	case EM_GETLINE: {
			int lineStart = doc.LineStart(wParam);
			int lineEnd = doc.LineStart(wParam + 1);
			char *ptr = reinterpret_cast<char *>(lParam);
			int iPlace = 0;
			for (int iChar = lineStart; iChar < lineEnd; iChar++)
				ptr[iPlace++] = doc.CharAt(iChar);
			ptr[iPlace] = '\0';
			return iPlace;
		}
		break;

	case EM_GETLINECOUNT:
		return doc.LinesTotal();

	case EM_GETMODIFY:
		return !doc.IsSavePoint();

	case EM_SETMODIFY:
		// Not really supported now that there is the save point stuff
		//doc.isModified = wParam;
		//return doc.isModified;


		return false;

	case EM_GETRECT:
		*(reinterpret_cast<PRectangle *>(lParam)) = GetClientRectangle();
		break;

	case EM_GETSEL:
	case EM_EXGETSEL:
		if (wParam)
			*reinterpret_cast<int *>(wParam) = SelectionStart();
		if (lParam)
			*reinterpret_cast<int *>(lParam) = SelectionEnd();
		return MAKELONG(SelectionStart(), SelectionEnd());
		break;

	case EM_SETSEL:
	case EM_EXSETSEL: {
			int nStart = static_cast<int>(wParam);
			int nEnd = static_cast<int>(lParam);
			if (nEnd < 0)
				nEnd = doc.Length();
			if (nStart < 0)
				nStart = nEnd;
			SetSelection(nEnd, nStart);
			EnsureCaretVisible();
			Redraw();
		}
		break;

	case EM_GETSELTEXT: {
			char *ptr = reinterpret_cast<char *>(lParam);
			int iPlace = 0;
			for (int iChar = SelectionStart(); iChar < SelectionEnd(); iChar++)
				ptr[iPlace++] = doc.CharAt(iChar);
			ptr[iPlace] = '\0';
			return iPlace;
		}
		break;

	case EM_GETWORDBREAKPROC:
		return 0;

	case EM_SETWORDBREAKPROC:
		break;

	case EM_LIMITTEXT:
		// wParam holds the number of characters control should be limited to
		break;

	case EM_GETLIMITTEXT:
		return 0xffffffff;

	case EM_GETOLEINTERFACE:
		return 0;

	case EM_LINEFROMCHAR:
	case EM_EXLINEFROMCHAR:
		return doc.LineFromPosition(wParam);

	case EM_LINEINDEX:
		return doc.LineStart(wParam);

	case EM_LINELENGTH:
		return doc.LineStart(wParam + 1) - doc.LineStart(wParam);

	case EM_REPLACESEL: {
			doc.BeginUndoAction();
			ClearSelection();
			char *replacement = reinterpret_cast<char *>(lParam);
			doc.InsertString(currentPos, replacement);
			doc.EndUndoAction();
			SetEmptySelection(currentPos + strlen(replacement));
			EnsureCaretVisible();
			Redraw();
		}
		break;

	case EM_LINESCROLL:
		ScrollTo(topLine + lParam);
		HorizontalScrollTo(xOffset + wParam * vs.spaceWidth);
		return TRUE;

	case EM_SCROLLCARET:
		EnsureCaretVisible();
		break;

	case EM_SETREADONLY:
		doc.SetReadOnly(wParam);
		return TRUE;

	case EM_SETRECT:
		break;

	case EM_CANPASTE:
		return 1;

	case EM_CHARFROMPOS: {
			Point *ppt = reinterpret_cast<Point *>(lParam);
			return PositionFromLocation(*ppt);
		}
		break;

	case EM_POSFROMCHAR: {
			Point *ppt = reinterpret_cast<Point *>(lParam);
			*ppt = LocationFromPosition(wParam);
			return 0;
		}
		break;

	case EM_FINDTEXT:
		return FindText(iMessage, wParam, lParam);
		break;

	case EM_FINDTEXTEX:
		return FindText(iMessage, wParam, lParam);
		break;

	case EM_GETTEXTRANGE: {
			TEXTRANGE *tr = reinterpret_cast<TEXTRANGE *>(lParam);
			doc.GetCharRange(tr->lpstrText, tr->chrg.cpMin, tr->chrg.cpMax - tr->chrg.cpMin);
			tr->lpstrText[tr->chrg.cpMax - tr->chrg.cpMin] = '\0';
			return tr->chrg.cpMax - tr->chrg.cpMin;
		}
		break;

	case EM_SELECTIONTYPE:
		if (currentPos == anchor)
			return SEL_EMPTY;
		else
			return SEL_TEXT;

	case EM_HIDESELECTION:
		hideSelection = wParam;
		Redraw();
		break;

	case EM_FORMATRANGE:
		return FormatRange(wParam, reinterpret_cast<FORMATRANGE *>(lParam));

		// Control specific mesages

	case SCI_ADDTEXT: {
			doc.InsertString(CurrentPosition(), reinterpret_cast<char *>(lParam), wParam);
			SetEmptySelection(currentPos + wParam);
			Redraw();
			return 0;
		}
		break;

	case SCI_ADDSTYLEDTEXT: {
			doc.InsertStyledString(CurrentPosition() * 2, reinterpret_cast<char *>(lParam), wParam);
			SetEmptySelection(currentPos + wParam / 2);
			Redraw();
			return 0;
		}
		break;

	case SCI_INSERTTEXT: {
			int insertPos = wParam;
			if (static_cast<short>(wParam) == -1)
				insertPos = CurrentPosition();
			int newCurrent = CurrentPosition();
			int newAnchor = anchor;
			char *sz = reinterpret_cast<char *>(lParam);
			doc.InsertString(insertPos, sz);
			if (newCurrent > insertPos)
				newCurrent += strlen(sz);
			if (newAnchor > insertPos)
				newAnchor += strlen(sz);
			SetEmptySelection(newCurrent);
			Redraw();
			return 0;
		}
		break;

	case SCI_CLEARALL:
		ClearAll();
		return 0;

	case SCI_SETUNDOCOLLECTION:
		doc.SetUndoCollection(static_cast<enum undoCollectionType>(wParam));
		return 0;

	case SCI_APPENDUNDOSTARTACTION:
		doc.AppendUndoStartAction();
		return 0;

	case SCI_BEGINUNDOACTION:
		doc.BeginUndoAction();
		return 0;

	case SCI_ENDUNDOACTION:
		doc.EndUndoAction();
		return 0;

	case SCI_GETCARETPERIOD:
		return caret.period;

	case SCI_SETCARETPERIOD:
		caret.period = wParam;
		break;

	case SCI_SETWORDCHARS: {
			doc.SetWordChars(reinterpret_cast<unsigned char *>(lParam));
		}
		break;

	case SCI_GETLENGTH:
		return doc.Length();

	case SCI_GETCHARAT:
		return doc.CharAt(wParam);

	case SCI_GETCURRENTPOS:
		return currentPos;

	case SCI_GETANCHOR:
		return anchor;

	case SCI_GETSTYLEAT:
		if (static_cast<short>(wParam) >= doc.Length())
			return 0;
		else
			return doc.StyleAt(wParam);

	case SCI_REDO:
		Redo();
		break;

	case SCI_SELECTALL:
		SelectAll();
		break;

	case SCI_SETSAVEPOINT:
		doc.SetSavePoint();
		NotifySavePoint(true);
		break;

	case SCI_GETSTYLEDTEXT: {
			TEXTRANGE *tr = reinterpret_cast<TEXTRANGE *>(lParam);
			int iPlace = 0;
			for (int iChar = tr->chrg.cpMin; iChar < tr->chrg.cpMax; iChar++) {
				tr->lpstrText[iPlace++] = doc.CharAt(iChar);
				tr->lpstrText[iPlace++] = doc.StyleAt(iChar);
			}
			tr->lpstrText[iPlace] = '\0';
			tr->lpstrText[iPlace + 1] = '\0';
			return iPlace;
		}
		break;

	case SCI_CANREDO:
		return doc.CanRedo() ? TRUE : FALSE;

	case SCI_MARKERLINEFROMHANDLE:
		return doc.LineFromHandle(wParam);

	case SCI_MARKERDELETEHANDLE:
		doc.DeleteMarkFromHandle(wParam);
		break;

	case SCI_GETVIEWWS:
		return vs.viewWhitespace;

	case SCI_SETVIEWWS:
		vs.viewWhitespace = wParam;
		Redraw();
		break;

	case SCI_GOTOLINE:
		GoToLine(wParam);
		break;

	case SCI_GOTOPOS:
		SetPosition(wParam);
		Redraw();
		break;

	case SCI_SETANCHOR:
		SetSelection(currentPos, wParam);
		//Platform::DebugPrintf("SetAnchor %d %d\n", currentPos, anchor);
		break;

	case SCI_GETCURLINE: {
			int lineCurrentPos = doc.LineFromPosition(currentPos);
			int lineStart = doc.LineStart(lineCurrentPos);
			unsigned int lineEnd = doc.LineStart(lineCurrentPos + 1);
			char *ptr = reinterpret_cast<char *>(lParam);
			unsigned int iPlace = 0;
			for (unsigned int iChar = lineStart; iChar < lineEnd && iPlace < wParam; iChar++)
				ptr[iPlace++] = doc.CharAt(iChar);
			ptr[iPlace++] = '\0';
			return currentPos - lineStart;
		}
		break;

	case SCI_GETENDSTYLED:
		return doc.GetEndStyled();

	case SCI_GETEOLMODE:
		return doc.eolMode;

	case SCI_SETEOLMODE:
		doc.eolMode = wParam;
		break;

	case SCI_STARTSTYLING:
		doc.StartStyling(wParam, lParam);
		break;

	case SCI_SETSTYLING:
		doc.SetStyleFor(wParam, lParam);
		break;

	case SCI_SETSTYLINGEX:  // Specify a complete styling buffer


		doc.SetStyles(wParam, reinterpret_cast<char *>(lParam));
		break;

	case SCI_SETMARGINWIDTH:
		if (wParam < 100) {
			vs.selMarginWidth = wParam;
		}
		InvalidateStyleRedraw();
		break;

	case SCI_SETBUFFEREDDRAW:
		bufferedDraw = wParam;
		break;

	case SCI_SETTABWIDTH:
		if (wParam > 0)
			doc.tabInChars = wParam;
		InvalidateStyleRedraw();
		break;

	case SCI_SETCODEPAGE:
		doc.dbcsCodePage = wParam;
		break;

	case SCI_SETLINENUMBERWIDTH:
		if (wParam < 200) {
			vs.lineNumberWidth = wParam;
		}
		InvalidateStyleRedraw();
		break;

	case SCI_SETUSEPALETTE:
		palette.allowRealization = wParam;
		InvalidateStyleRedraw();
		break;

		// Marker definition and setting
	case SCI_MARKERDEFINE:
		if (wParam <= MARKER_MAX)
			vs.markers[wParam].markType = lParam;
		InvalidateStyleData();
		RedrawSelMargin();
		break;
	case SCI_MARKERSETFORE:
		if (wParam <= MARKER_MAX)
			vs.markers[wParam].fore.desired = Colour(lParam);
		InvalidateStyleData();
		RedrawSelMargin();
		break;
	case SCI_MARKERSETBACK:
		if (wParam <= MARKER_MAX)
			vs.markers[wParam].back.desired = Colour(lParam);
		InvalidateStyleData();
		RedrawSelMargin();
		break;
	case SCI_MARKERADD: {
			int markerID = doc.AddMark(wParam, lParam);
			RedrawSelMargin();
			return markerID;
		}
		break;

	case SCI_MARKERDELETE:
		doc.DeleteMark(wParam, lParam);
		RedrawSelMargin();
		break;

	case SCI_MARKERDELETEALL:
		doc.DeleteAllMarks(static_cast<int>(wParam));
		RedrawSelMargin();
		break;

	case SCI_MARKERGET:
		return doc.GetMark(wParam);

	case SCI_MARKERNEXT: {
			for (int iLine = wParam; iLine < doc.LinesTotal(); iLine++) {
				if ((doc.GetMark(iLine) & lParam) != 0)
					return iLine;
			}
		}
		return - 1;

	case SCI_STYLECLEARALL:
		vs.ClearStyles();
		InvalidateStyleRedraw();
		break;

	case SCI_STYLESETFORE:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].fore.desired = Colour(lParam);
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETBACK:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].back.desired = Colour(lParam);
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETBOLD:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].bold = lParam;
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETITALIC:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].italic = lParam;
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETEOLFILLED:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].eolFilled = lParam;
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETSIZE:
		if (wParam <= STYLE_MAX) {
			vs.styles[wParam].size = lParam;
			InvalidateStyleRedraw();
		}
		break;
	case SCI_STYLESETFONT:
		if (wParam <= STYLE_MAX) {
			strcpy(vs.styles[wParam].fontName, reinterpret_cast<char *>(lParam));
			InvalidateStyleRedraw();
		}
		break;

	case SCI_SETFORE:
		vs.foreground.desired = Colour(wParam);
		InvalidateStyleRedraw();
		break;

	case SCI_SETBACK:
		vs.background.desired = Colour(wParam);
		InvalidateStyleRedraw();
		break;

	case SCI_SETBOLD:
		vs.bold = wParam;
		InvalidateStyleRedraw();
		break;

	case SCI_SETITALIC:
		vs.italic = wParam;
		InvalidateStyleRedraw();
		break;

	case SCI_SETSIZE:
		vs.size = wParam;
		InvalidateStyleRedraw();
		break;

	case SCI_SETFONT:
		strcpy(vs.fontName, reinterpret_cast<char *>(wParam));
		InvalidateStyleRedraw();
		break;

	case SCI_SETSELFORE:
		vs.selforeset = wParam;
		vs.selforeground.desired = Colour(lParam);
		InvalidateStyleRedraw();
		break;

	case SCI_SETSELBACK:
		vs.selbackset = wParam;
		vs.selbackground.desired = Colour(lParam);
		InvalidateStyleRedraw();
		break;

	case SCI_SETCARETFORE:
		vs.caretcolour.desired = Colour(wParam);
		InvalidateStyleRedraw();
		break;

	case SCI_ASSIGNCMDKEY:
		kmap.AssignCmdKey(LOWORD(wParam), HIWORD(wParam), lParam);
		break;

	case SCI_CLEARCMDKEY:
		kmap.AssignCmdKey(LOWORD(wParam), HIWORD(wParam), WM_NULL);
		break;

	case SCI_CLEARALLCMDKEYS:
		kmap.Clear();
		break;

	case SCI_INDICSETSTYLE:
		if (wParam <= INDIC_MAX) {
			vs.indicators[wParam].style = lParam;
			InvalidateStyleRedraw();
		}
		break;

	case SCI_INDICGETSTYLE:
		return (wParam <= INDIC_MAX) ? vs.indicators[wParam].style : 0;

	case SCI_INDICSETFORE:
		if (wParam <= INDIC_MAX) {
			vs.indicators[wParam].fore.desired = Colour(lParam);
			InvalidateStyleRedraw();
		}
		break;

	case SCI_INDICGETFORE:
		return (wParam <= INDIC_MAX) ? vs.indicators[wParam].fore.desired.AsLong() : 0;

	case SCI_LINEDOWN:
	case SCI_LINEDOWNEXTEND:
	case SCI_LINEUP:
	case SCI_LINEUPEXTEND:
	case SCI_CHARLEFT:
	case SCI_CHARLEFTEXTEND:
	case SCI_CHARRIGHT:
	case SCI_CHARRIGHTEXTEND:
	case SCI_WORDLEFT:
	case SCI_WORDLEFTEXTEND:
	case SCI_WORDRIGHT:
	case SCI_WORDRIGHTEXTEND:
	case SCI_HOME:
	case SCI_HOMEEXTEND:
	case SCI_LINEEND:
	case SCI_LINEENDEXTEND:
	case SCI_DOCUMENTSTART:
	case SCI_DOCUMENTSTARTEXTEND:
	case SCI_DOCUMENTEND:
	case SCI_DOCUMENTENDEXTEND:
	case SCI_PAGEUP:
	case SCI_PAGEUPEXTEND:
	case SCI_PAGEDOWN:
	case SCI_PAGEDOWNEXTEND:
	case SCI_EDITTOGGLEOVERTYPE:
	case SCI_CANCEL:
	case SCI_DELETEBACK:
	case SCI_TAB:
	case SCI_BACKTAB:
	case SCI_NEWLINE:
	case SCI_FORMFEED:
	case SCI_VCHOME:
	case SCI_VCHOMEEXTEND:
	case SCI_ZOOMIN:
	case SCI_ZOOMOUT:
	case SCI_DELWORDLEFT:
	case SCI_DELWORDRIGHT:
		return KeyCommand(iMessage);

	default:
		return DefWndProc(iMessage, wParam, lParam);
	}
	//Platform::DebugPrintf("end wnd proc\n");
	return 0l;
}
