// SciTE - Scintilla based Text Editor
// PropSet.cxx - a java style properties file module
// Copyright 1998-1999 by Neil Hodgson <neilh@hare.net.au>
// The License.txt file describes the conditions under which this software may be distributed.

// Maintain a dictionary of properties

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include "Platform.h"

#include "PropSet.h"

bool EqualCaseInsensitive(const char *a, const char *b) {
#ifdef PLAT_GTK
	return 0 == strcasecmp(a, b);
#else
	return 0 == stricmp(a, b);
#endif
}

// Get a line of input. If end of line escaped with '\\' then continue reading.
static bool GetFullLine(const char *&fpc, int &lenData, char *s, int len) {
	bool continuation = true;
	while ((len > 1) && lenData > 0) {
		char ch = *fpc;
		fpc++;
		lenData--;
		if ((ch == '\r') || (ch == '\n')) {
			if (!continuation) {
				if ((lenData > 0) && (ch == '\r') && ((*fpc) == '\n')) {
					// munch the second half of a crlf
					fpc++;
					lenData--;
				}
				*s++ = '\0';
				return true;
			}
		} else if ((ch == '\\') && (lenData > 0) && ((*fpc== '\r') || (*fpc== '\n'))) {
			continuation = true;
		} else {
			continuation = false;
			*s++ = ch;
			len--;
		}
	}
	return false;
}

PropSet::PropSet() {
	superPS = 0;
	size = 10;
	used = 0;
	vals = new char*[size];
}

PropSet::~PropSet() {
	superPS = 0;
	Clear();
	delete []vals;
}

void PropSet::EnsureCanAddEntry() {
	if (used >= size - 2) {
		int newsize = size + 10;
		char **newvals = new char*[newsize];

		for (int i=0;i<used;i++) {
			newvals[i] = vals[i];
		}
		delete [] vals;
		vals = newvals;
		size = newsize;
	}
}

void PropSet::Set(const char *key, const char *val) {
	EnsureCanAddEntry();
	for (int i=0;i<used;i+=2) {
		if (EqualCaseInsensitive(vals[i], key)) {
			// Replace current value
			free(vals[i+1]);
			vals[i+1] = strdup(val);
			return;
		}
	}
	// Not found
	vals[used++] = strdup(key);
	vals[used++] = strdup(val);
}

void PropSet::Set(char *keyval) {
	char *eqat = strchr(keyval, '=');
	if (eqat) {
		*eqat = '\0';
		Set(keyval, eqat + 1);
		*eqat = '=';
	}
}

SString PropSet::Get(const char *key) {
	for (int i=0;i<used;i+=2) {
		if (EqualCaseInsensitive(vals[i], key)) {
			return vals[i+1];
		}
	}
	if (superPS) {
		// Failed here, so try in base property set
		return superPS->Get(key);
	} else {
		return "";
	}
}

int PropSet::GetInt(const char *key) {
	return Get(key).value();
}

bool isprefix(const char *target, const char *prefix) {
	while (*target && *prefix) {
		if (toupper(*target) != toupper(*prefix))
			return false;
		target++;
		prefix++;
	}
	if (*prefix)
		return false;
	else
		return true;
}

bool issuffix(const char *target, const char *suffix) {
	int lentarget = strlen(target);
	int lensuffix = strlen(suffix);
	if (lensuffix > lentarget)
		return false;
	for (int i=lensuffix-1; i >= 0; i--) {
		if (toupper(target[i + lentarget - lensuffix]) != toupper(suffix[i]))
			return false;
	}
	return true;
}

SString PropSet::GetWild(const char *keybase, const char *filename) {
	int lenbase = strlen(keybase);
	int lenfile = strlen(filename);
	for (int i=0;i<used;i+=2) {
		if (isprefix(vals[i], keybase)) {
			const char *keyfile = vals[i] + strlen(keybase);
			if (*keyfile == '*') {
				if (issuffix(filename, keyfile + 1)) {
					return vals[i+1];
				}
			} else if (EqualCaseInsensitive(keyfile, filename)) {
				return vals[i+1];
			} else if (EqualCaseInsensitive(vals[i], keybase)) {
				return vals[i+1];
			}
		}
	}
	if (superPS) {
		// Failed here, so try in base property set
		return superPS->GetWild(keybase, filename);
	} else {
		return "";
	}
}

SString PropSet::GetNewExpand(const char *keybase, const char *filename) {
	char *base = strdup(GetWild(keybase, filename).c_str());
	char *cpvar = strstr(base, "$(");
	while (cpvar) {
		char *cpendvar = strchr(cpvar, ')');
		if (cpendvar) {
			int lenvar = cpendvar - cpvar - 2;	// Subtract the $()
			char *var = static_cast<char *>(malloc(lenvar + 1));
			strncpy(var, cpvar+2, lenvar);
			var[lenvar] = '\0';
			SString val = GetWild(var, filename);
			int newlenbase = strlen(base) + val.length() - lenvar;
			char *newbase = static_cast<char *>(malloc(newlenbase));
			strncpy(newbase, base, cpvar - base);
			strcpy(newbase + (cpvar - base), val.c_str());
			strcpy(newbase + (cpvar - base) + val.length(), cpendvar + 1);
			free(var);
			free(base);
			base = newbase;
		}
		cpvar = strstr(base, "$(");
	}
	SString sret = base;
	free(base);
	return sret;
}

void PropSet::Clear() {
	for (int i=0;i<used;i++) {
		free(vals[i]);
		vals[i] = 0;
	}
	used = 0;
}

void PropSet::ReadFromMemory(const char *data, int len) {
	if (len > 0) {
		const char *pd = data;
		char linebuf[4000];
		while (GetFullLine(pd, len, linebuf, sizeof(linebuf))) {
			if (isalpha(linebuf[0]))
				Set(linebuf);
		}
	}
}

void PropSet::Read(const char *filename) {
	//printf("Opening properties <%s>\n", filename);
	Clear();
	char propsData[60000];
	FILE *rcfile = fopen(filename, "rb");
	if (rcfile) {
		int lenFile = fread(propsData,1,sizeof(propsData),rcfile);
		fclose(rcfile);
		ReadFromMemory(propsData, lenFile);
	} else {
		//printf("Could not open <%s>\n", filename);
	}
}

